/**
 * CMI : Cluster Method Invocation
 * Copyright (C) 2008 Bull S.A.S.
 * Contact: cmi@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 * --------------------------------------------------------------------------
 * $Id: ClusterView.java 2122 2008-09-27 09:41:35Z loris $
 * --------------------------------------------------------------------------
 */

package org.ow2.cmi.controller.server;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.jcip.annotations.ThreadSafe;

import org.ow2.cmi.controller.server.filter.IFilter;
import org.ow2.cmi.reference.CMIReference;
import org.ow2.cmi.reference.ObjectNotFoundException;
import org.ow2.cmi.reference.ServerNotFoundException;
import org.ow2.cmi.reference.ServerRef;

/**
 * @author Loris Bouzonnet
 */
@ThreadSafe
public class ClusterView implements Serializable {

    /**
     *
     */
    private static final long serialVersionUID = -5710461778454543508L;

    private final Map<String, ClusteredObjectView> clusteredObjectsViews;

    private final Map<String, ServerView> serverViews;

    private final List<IFilter> filters;

    private int delayToRefresh;

    public ClusterView() {
        clusteredObjectsViews = new HashMap<String, ClusteredObjectView>();
        serverViews = new HashMap<String, ServerView>();
        filters = Collections.synchronizedList(new ArrayList<IFilter>());
    }

    public void clear() {
        clusteredObjectsViews.clear();
        filters.clear();
        delayToRefresh = -1;
    }

    public boolean containsObject(final String objectName) {
        synchronized (clusteredObjectsViews) {
            return clusteredObjectsViews.containsKey(objectName);
        }
    }

    public DistributedObjectInfo getObjectInfo(final String objectName) throws ObjectNotFoundException {
        synchronized (clusteredObjectsViews) {
            ClusteredObjectView clusteredObjectView =
                clusteredObjectsViews.get(objectName);
            if(clusteredObjectView == null) {
                throw new ObjectNotFoundException(
                        "Cannot get object infos", objectName);
            }
            return clusteredObjectView.getObjectInfo();
        }
    }

    public Set<CMIReference> getCMIReferences(final String objectName,
            final String protocolName) throws ObjectNotFoundException {
        synchronized (clusteredObjectsViews) {
            ClusteredObjectView clusteredObjectView =
                clusteredObjectsViews.get(objectName);
            if(clusteredObjectView == null) {
                throw new ObjectNotFoundException(
                    "Cannot get CMI references", objectName);
            }
            return clusteredObjectView.getCMIReferences(protocolName);
        }
    }


    public List<CMIReference> getCMIReferences(final String objectName) throws ObjectNotFoundException {
        synchronized (clusteredObjectsViews) {
            ClusteredObjectView clusteredObjectView =
                clusteredObjectsViews.get(objectName);
            if(clusteredObjectView == null) {
                throw new ObjectNotFoundException(
                    "Cannot get CMI references", objectName);
            }
            return clusteredObjectView.getCMIReferences();
        }
    }

    public Set<String> getObjectNames() {
        synchronized (clusteredObjectsViews) {
            return new HashSet<String>(clusteredObjectsViews.keySet());
        }
    }

    public DistributedObjectInfo addObjectInfo(
            final DistributedObjectInfo distributedObjectInfo, final boolean override) {
        DistributedObjectInfo oldDistributedObjectInfo;
        synchronized (clusteredObjectsViews) {
            ClusteredObjectView clusteredObjectView =
                clusteredObjectsViews.get(distributedObjectInfo.getObjectName());
            if(clusteredObjectView == null) {
                clusteredObjectView = new ClusteredObjectView();
                clusteredObjectView.setObjectInfo(distributedObjectInfo);
                clusteredObjectsViews.put(distributedObjectInfo.getObjectName(), clusteredObjectView);
            } else if(override) {
                oldDistributedObjectInfo = clusteredObjectView.getObjectInfo();
                if(!oldDistributedObjectInfo.equals(distributedObjectInfo)) {
                    clusteredObjectView.setObjectInfo(distributedObjectInfo);
                }
                return oldDistributedObjectInfo;
            }
        }
        return null;
    }

    public void addCMIReference(final CMIReference cmiReference) {
        synchronized (clusteredObjectsViews) {
            ClusteredObjectView clusteredObjectView =
                clusteredObjectsViews.get(cmiReference.getObjectName());
            if(clusteredObjectView == null) {
                clusteredObjectView = new ClusteredObjectView();
                clusteredObjectsViews.put(cmiReference.getObjectName(), clusteredObjectView);
            }
            clusteredObjectView.addCMIReference(cmiReference);
        }
    }

    public boolean removeCMIReference(final CMIReference cmiReference) throws ObjectNotFoundException {
        synchronized (clusteredObjectsViews) {
            ClusteredObjectView clusteredObjectView =
                clusteredObjectsViews.get(cmiReference.getObjectName());
            if(clusteredObjectView == null) {
                throw new ObjectNotFoundException(
                        "Cannot get CMI references",
                        cmiReference.getObjectName());
            }
            return clusteredObjectView.removeCMIReference(cmiReference);
        }
    }

    public void setDelay(final int delay) {
        this.delayToRefresh = delay;
    }

    public int getDelay() {
        return delayToRefresh;
    }

    public List<IFilter> getFilters() {
        return Collections.unmodifiableList(filters);
    }

    public void addFilter(final IFilter filter) {
        filters.add(filter);
    }

    public boolean removeFilter(final IFilter filter) {
        return filters.remove(filter);
    }

    public boolean addServerToBlackList(final ServerRef serverRef) {
        boolean blacklisted = false;
        synchronized (serverViews) {
            ServerView serverView = serverViews.get(serverRef.getProviderURL());
            if(serverView == null) {
                serverView = new ServerView();
                serverViews.put(serverRef.getProviderURL(), serverView);
            } else {
                blacklisted = serverView.isBlacklisted();
            }
            if(!blacklisted) {
                serverView.setBlacklisted(true);
            }
        }
        return blacklisted;
    }

    public boolean isServerBlackListed(final ServerRef serverRef) {
        synchronized (serverViews) {
            ServerView serverView = serverViews.get(serverRef.getProviderURL());
            if(serverView == null) {
                return false;
            }
            return serverView.isBlacklisted();
        }
    }

    public boolean removeServerFromBlackList(final ServerRef serverRef) throws ServerNotFoundException {
        boolean blacklisted = false;
        synchronized (serverViews) {
            ServerView serverView = serverViews.get(serverRef.getProviderURL());
            if(serverView == null) {
                throw new ServerNotFoundException(
                        "Cannot remove from the blacklist an unkwnown server", serverRef);
            }
            blacklisted = serverView.isBlacklisted();
            if(blacklisted) {
                serverView.setBlacklisted(false);
            }
        }
        return blacklisted;
    }

    public Set<String> getClusterNames() {
        synchronized (clusteredObjectsViews) {
            Set<String> clusterNames = new HashSet<String>();
            for(String objectName : clusteredObjectsViews.keySet()) {
                clusterNames.add(clusteredObjectsViews.get(objectName).getObjectInfo().getClusterName());
            }
            return clusterNames;
        }
    }

    public Set<String> getObjectNames(final String clusterName) {
        synchronized (clusteredObjectsViews) {
            Set<String> objectNames = new HashSet<String>();
            for(String objectName : clusteredObjectsViews.keySet()) {
                if(clusteredObjectsViews.get(objectName).getObjectInfo().getClusterName().equals(clusterName)) {
                    objectNames.add(objectName);
                }
            }
            return objectNames;
        }
    }

    public int setLoadFactor(final ServerRef serverRef, final int loadFactor) {
        int oldLoadFactor;
        synchronized (serverViews) {
            ServerView serverView = serverViews.get(serverRef.getProviderURL());
            if(serverView == null) {
                serverView = new ServerView();
                serverViews.put(serverRef.getProviderURL(), serverView);
                oldLoadFactor = -1;
            } else {
                oldLoadFactor = serverView.getLoadFactor();
            }
            if(oldLoadFactor != loadFactor) {
                serverView.setLoadFactor(loadFactor);
            }
        }
        return oldLoadFactor;
    }

    public int getLoadFactor(final ServerRef serverRef) throws ServerNotFoundException {
        synchronized (serverViews) {
            ServerView serverView = serverViews.get(serverRef.getProviderURL());
            if(serverView == null) {
                throw new ServerNotFoundException(
                        "Cannot remove from the blacklist an unkwnown server", serverRef);
            }
            return serverView.getLoadFactor();
        }
    }

    public Map<String, ? extends ServerView> getServerViews() {
        return serverViews;
    }

    public void removeClusteredObjects(final String providerURL) {
        synchronized (clusteredObjectsViews) {
            // Remove its objects
            for(ClusteredObjectView clusteredObjectView : clusteredObjectsViews.values()) {
                clusteredObjectView.removeServerRef(providerURL);
            }
        }
    }

}
